@@ -30,6 +30,7 @@
 #include "grab_key.h"
 
 
+#define STR_KEY_LEN 1000
 
 int nb_keys;
 Keys_t *keys;
@@ -59,14 +60,19 @@ close_keys (void)
     free_key (&keys[i]);
 
   if (keys != NULL)
-    free (keys);
+    {
+      free (keys);
+      keys = NULL;
+    }
+
+  nb_keys = 0;
 }
 
 
 
 int
 add_key (KeyType_t type, EventType_t event_type, KeySym keysym, KeyCode keycode,
-	 unsigned int button, unsigned int modifier, char *command)
+	 unsigned int button, unsigned int modifier, char *command, SCM function)
 {
   Keys_t *keys_bis = NULL;
   int i;
@@ -102,15 +108,15 @@ add_key (KeyType_t type, EventType_t event_type, KeySym keysym, KeyCode keycode,
   /* set key */
   if (type == SYM)
     {
-      set_keysym (&keys[nb_keys], event_type, keysym, modifier, command);
+      set_keysym (&keys[nb_keys], event_type, keysym, modifier, command, function);
     }
   else if (type == BUTTON)
     {
-      set_button (&keys[nb_keys], event_type, button, modifier, command);
+      set_button (&keys[nb_keys], event_type, button, modifier, command, function);
     }
   else
     {
-      set_keycode (&keys[nb_keys], event_type, keycode, modifier, command);
+      set_keycode (&keys[nb_keys], event_type, keycode, modifier, command, function);
     }
 
   /* new key */
@@ -121,6 +127,9 @@ add_key (KeyType_t type, EventType_t event_type, KeySym keysym, KeyCode keycode,
 
 
 
+
+
+
 void
 show_key_binding (Display * d)
 {
@@ -147,51 +156,51 @@ modifier_to_string (unsigned int modifier, char *str)
   if (modifier & ControlMask)
     {
       if (str[0])
-	strcat (str, "+");
-      strcat (str, modifier_string[0]);
+	strncat (str, "+", STR_KEY_LEN);
+      strncat (str, modifier_string[0], STR_KEY_LEN);
     }
 
   if (modifier & ShiftMask)
     {
       if (str[0])
-	strcat (str, "+");
-      strcat (str, modifier_string[1]);
+	strncat (str, "+", STR_KEY_LEN);
+      strncat (str, modifier_string[1], STR_KEY_LEN);
     }
 
 
   if (modifier & Mod1Mask)
     {
       if (str[0])
-	strcat (str, "+");
-      strcat (str, modifier_string[2]);
+	strncat (str, "+", STR_KEY_LEN);
+      strncat (str, modifier_string[2], STR_KEY_LEN);
     }
 
   if (modifier & Mod2Mask)
     {
       if (str[0])
-	strcat (str, "+");
-      strcat (str, modifier_string[3]);
+	strncat (str, "+", STR_KEY_LEN);
+      strncat (str, modifier_string[3], STR_KEY_LEN);
     }
 
   if (modifier & Mod3Mask)
     {
       if (str[0])
-	strcat (str, "+");
-      strcat (str, modifier_string[4]);
+	strncat (str, "+", STR_KEY_LEN);
+      strncat (str, modifier_string[4], STR_KEY_LEN);
     }
 
   if (modifier & Mod4Mask)
     {
       if (str[0])
-	strcat (str, "+");
-      strcat (str, modifier_string[5]);
+	strncat (str, "+", STR_KEY_LEN);
+      strncat (str, modifier_string[5], STR_KEY_LEN);
     }
 
   if (modifier & Mod5Mask)
     {
       if (str[0])
-	strcat (str, "+");
-      strcat (str, modifier_string[6]);
+	strncat (str, "+", STR_KEY_LEN);
+      strncat (str, modifier_string[6], STR_KEY_LEN);
     }
 }
 
@@ -199,7 +208,7 @@ modifier_to_string (unsigned int modifier, char *str)
 void
 print_key (Display * d, Keys_t * key)
 {
-  char str[1000];
+  char str[STR_KEY_LEN];
 
   if (verbose)
     {
@@ -208,21 +217,21 @@ print_key (Display * d, Keys_t * key)
 	  modifier_to_string (key->modifier, str);
 
 	  printf ("\"%s\"\n    %s%s%s%s\n",
-		  key->command != NULL ? key->command : "NoCommand",
+		  key->command != NULL ? key->command : ( key->function != 0 ? "(Scheme function)" : "NoCommand" ),
 		  key->event_type == PRESS ? "" : "Release + ",
 		  str, str[0] ? " + " : "", XKeysymToString (key->key.sym));
 	}
       else if (key->type == BUTTON)
 	{
 	  printf ("\"%s\"\n    %sm:0x%x + b:%d   (mouse)\n",
-		  key->command != NULL ? key->command : "NoCommand",
+		  key->command != NULL ? key->command : ( key->function != 0 ? "(Scheme function)" : "NoCommand" ),
 		  key->event_type == PRESS ? "" : "Release + ",
 		  key->modifier, key->key.button);
 	}
       else
 	{
 	  printf ("\"%s\"\n    %sm:0x%x + c:%d\n",
-		  key->command != NULL ? key->command : "NoCommand",
+		  key->command != NULL ? key->command : ( key->function != 0 ? "(Scheme function)" : "NoCommand" ),
 		  key->event_type == PRESS ? "" : "Release + ",
 		  key->modifier, key->key.code);
 	  if (d != NULL)
@@ -247,48 +256,75 @@ print_key (Display * d, Keys_t * key)
 
 void
 set_keysym (Keys_t * key, EventType_t event_type, KeySym keysym,
-	    unsigned int modifier, char *command)
+	    unsigned int modifier, char *command, SCM function)
 {
   key->type = SYM;
   key->event_type = event_type;
   key->key.sym = keysym;
   key->modifier = modifier;
 
-  key->command = (char *) malloc ((strlen (command) + 1) * sizeof (char));
+  if (command != NULL)
+    {
+      key->command = (char *) malloc ((strlen (command) + 1) * sizeof (char));
 
-  if (key->command != NULL)
-    strcpy (key->command, command);
+      if (key->command != NULL)
+	strncpy (key->command, command, strlen (command) + 1);
+    }
+  else
+    {
+      key->command = NULL;
+    }
+
+  key->function = function;
 }
 
 void
-set_keycode (Keys_t * key, EventType_t event_type, KeySym keycode, 
-	     unsigned int modifier, char *command)
+set_keycode (Keys_t * key, EventType_t event_type, KeyCode keycode, 
+	     unsigned int modifier, char *command, SCM function)
 {
   key->type = CODE;
   key->event_type = event_type;
   key->key.code = keycode;
   key->modifier = modifier;
 
-  key->command = (char *) malloc ((strlen (command) + 1) * sizeof (char));
-
-  if (key->command != NULL)
-    strcpy (key->command, command);
+  if (command != NULL)
+    {
+      key->command = (char *) malloc ((strlen (command) + 1) * sizeof (char));
+      
+      if (key->command != NULL)
+	strncpy (key->command, command, strlen (command) + 1);
+    }
+  else
+    {
+      key->command = NULL;
+    }
+  
+  key->function = function;
 }
 
 
 void
 set_button (Keys_t * key, EventType_t event_type, unsigned int button, 
-	    unsigned int modifier, char *command)
+	    unsigned int modifier, char *command, SCM function)
 {
   key->type = BUTTON;
   key->event_type = event_type;
   key->key.button = button;
   key->modifier = modifier;
-
-  key->command = (char *) malloc ((strlen (command) + 1) * sizeof (char));
-
-  if (key->command != NULL)
-    strcpy (key->command, command);
+  
+  if (command != NULL)
+    {
+      key->command = (char *) malloc ((strlen (command) + 1) * sizeof (char));
+      
+      if (key->command != NULL)
+	strncpy (key->command, command, strlen (command) + 1);
+    }
+  else
+    {
+      key->command = NULL;
+    }
+  
+  key->function = function;
 }
 
 
@@ -304,24 +340,81 @@ free_key (Keys_t * key)
 
   if (key->command != NULL)
     free (key->command);
+
+  key->function = 0;
+}
+
+
+
+
+int
+remove_key (KeyType_t type, EventType_t event_type, KeySym keysym, KeyCode keycode,
+	    unsigned int button, unsigned int modifier)
+{
+  int i, found_index = -1;
+  Keys_t *keys_bis = NULL;
+
+  if (keys == NULL)
+    {
+      return (-1);
+    }
+  
+  modifier &= ~(numlock_mask | capslock_mask | scrolllock_mask);
+
+  for (i = 0; i < nb_keys; i++)
+    {
+      if (keys[i].type == type &&
+	  keys[i].event_type == event_type &&
+	  keys[i].modifier == modifier &&
+	  ((type == SYM && keys[i].key.sym == keysym) ||
+	   (type == CODE && keys[i].key.code == keycode) ||
+	   (type == BUTTON && keys[i].key.button == button)))
+	found_index = i;
+    }
+
+  if (found_index != -1)
+    {
+      if (verbose)
+	printf ("Removing key index %d\n", found_index);
+  
+      /* make new array keys_bis */
+      keys_bis = (Keys_t *) malloc ((nb_keys - 1) * sizeof (Keys_t));
+      if (keys_bis == NULL)
+	return (-1);
+
+      for (i = 0; i < found_index; i++)
+	keys_bis[i] = keys[i];
+      
+      for (i = found_index + 1; i < nb_keys; i++)
+	keys_bis[i - 1] = keys[i];
+
+      free_key (&keys[found_index]);
+      free (keys);
+
+      /* make keys_bis as keys */
+      keys = keys_bis;
+
+      nb_keys -= 1;
+    }
+
+  return (0);
 }
 
 
 
 
 
+
+
 void
-start_command_key (Keys_t * key)
+run_command (char *command)
 {
   pid_t pid;
   
-  if (key->command == NULL)
-    return;
-
 #ifdef FORK_FLAG  
   if (verbose)
     printf ("Start program with fork+exec call\n");
-
+  
   //  if (fork() == 0)
   //  execlp ("sh", "sh", "-c", key->command, NULL);
   if (!(pid = fork()))
@@ -329,7 +422,7 @@ start_command_key (Keys_t * key)
       setsid();
       switch (fork())
 	{
-	case 0: execlp ("sh", "sh", "-c", key->command, NULL);
+	case 0: execlp ("sh", "sh", "-c", command, (char *) NULL);
 	  break;
 	default: _exit(0);
 	  break;
@@ -341,6 +434,24 @@ start_command_key (Keys_t * key)
   if (verbose)
     printf ("Start program with system call\n");
   
-  system (key->command);
+  system (command);
 #endif
 }
+
+
+void
+start_command_key (Keys_t * key)
+{
+  if (key->command == NULL)
+    {
+#ifdef GUILE_FLAG
+      if (key->function != 0)
+	{
+	  scm_call_0 (key->function);
+	}
+#endif
+      return;
+    }
+
+  run_command (key->command);
+}
